@@ -3,18 +3,19 @@
 A Python interface to Graphviz.  
 
 """
-#    Copyright (C) 2006,2007 by 
+#    Copyright (C) 2006-2008 by 
 #    Aric Hagberg <hagberg@lanl.gov>
 #    Dan Schult <dschult@colgate.edu>
 #    Manos Renieris, http://www.cs.brown.edu/~er/
 #    Distributed with BSD license.     
 #    All rights reserved, see LICENSE for details.
-
-import graphviz as gv
-import UserDict
+import subprocess
 import sys
 import threading
 import warnings
+import UserDict
+
+import graphviz as gv
 
 class PipeReader(threading.Thread):
     """Read and write pipes using threads.
@@ -39,34 +40,31 @@ class _Action:
     find, create = 0,1
 
 class AGraph(object):
-    """
-    Class for Graphviz agraph type.
-
-    Example use::
-         
-      >>> G=AGraph()         
-      >>> G=AGraph(directed=True)
-
-      G=AGraph("file.dot")   
-
+    """Class for Graphviz agraph type.
+    
+    Example use
+    
+    >>> G=AGraph()         
+    >>> G=AGraph(directed=True)
+    >>> G=AGraph("file.dot")   
+    
     Graphviz graph keyword parameters are processed so you may add
     them like
-
+    
     >>> G=AGraph(landscape='true',ranksep='0.1')
-
+    
     or alternatively
-
+    
     >>> G=AGraph()
     >>> G.graph_attr.update(landscape='true',ranksep='0.1')
-
+    
     and
-
+    
     >>> G.node_attr.update(color='red')
     >>> G.edge_attr.update(len='2.0',color='blue')
-
+    
     See http://www.graphviz.org/doc/info/attrs.html
     for a list of attributes.
-    
 
     Keyword parameters:
 
@@ -81,18 +79,20 @@ class AGraph(object):
     >>> h=AGraph.handle
     >>> C=AGraph(h)
 
-    name is a string name for the graph        
+    Parameters:: 
 
-    strict=True|False (True for simple graphs)
+      name:    String name for the graph        
 
-    directed=True|False
+      strict: True|False (True for simple graphs)
 
-    data is a dictionary of dictionaries or dictionary of lists
-    representing nodes or edges to load into intial graph
+      directed: True|False
 
-    string is a string containing a dot format graph
+      data: Dictionary of dictionaries or dictionary of lists
+      representing nodes or edges to load into intial graph
 
-    handle is a Swig pointer to an agraph_t data structure
+      string:  String containing a dot format graph
+
+      handle:  Swig pointer to an agraph_t data structure
 
     """
     def __init__(self, thing=None, file=None, name=None,
@@ -164,6 +164,11 @@ class AGraph(object):
         # this is not graph isomorphism
         return self.string()==other.string()
 
+    def __hash__(self):
+        # hash the string representation for id
+        return hash(self.string())
+
+
     def __iter__(self):
         # provide "for n in G"
         return self.nodes_iter()
@@ -184,7 +189,7 @@ class AGraph(object):
 #        self.add_edge(u,v)
 
     def add_node(self, n, **attr):
-        """Add a single node n to the graph.
+        """Add a single node n.
 
         If n is not a string, conversion to a string will be attempted.
         String conversion will work if n has valid string representation
@@ -214,11 +219,11 @@ class AGraph(object):
             nh=gv.agnode(self.handle,n,_Action.find)
         except KeyError:
             nh=gv.agnode(self.handle,n,_Action.create)
-            nh=Node(self,n)
-            nh.attr.update(**attr)
+            node=Node(self,nh=nh)
+            node.attr.update(**attr)
 
     def add_nodes_from(self, nbunch, **attr):
-        """Add nodes to graph from a container nbunch.
+        """Add nodes from a container nbunch.
 
         nbunch can be any iterable container such as a list or dictionary
 
@@ -237,43 +242,47 @@ class AGraph(object):
         for n in nbunch:
             self.add_node(n,**attr)
 
-    def delete_node(self,n):
-        """Delete the single node n from graph.
+    def remove_node(self,n):
+        """Remove the single node n.
 
-        Attempting to delete a node that isn't in the graph will produce
+        Attempting to remove a node that isn't in the graph will produce
         an error.
         
         >>> G=AGraph()
         >>> G.add_node('a')
-        >>> G.delete_node('a')
+        >>> G.remove_node('a')
 
         """
         if not self._is_string_like(n):  n=str(n)
         try:
             nh=gv.agnode(self.handle,n,_Action.find)
-            gv.agdelnode(nh)
+            gv.agdelnode(self.handle,nh)
         except KeyError:
             raise KeyError("node %s not in graph"%n)
 
-    def delete_nodes_from(self,nbunch):
-        """Delete nodes in graph from a container nbunch.
+    delete_node=remove_node
+
+    def remove_nodes_from(self,nbunch):
+        """Remove nodes from a container nbunch.
 
         nbunch can be any iterable container such as a list or dictionary
 
         >>> G=AGraph()
         >>> nlist=['a','b',1,'spam']
         >>> G.add_nodes_from(nlist)
-        >>> G.delete_nodes_from(nlist)
+        >>> G.remove_nodes_from(nlist)
         """
         for n in nbunch: 
-            self.delete_node(n)
+            self.remove_node(n)
+
+    delete_nodes_from=remove_nodes_from
 
     def nodes_iter(self):
         """Return an iterator over all the nodes in the graph."""
         nh=gv.agfstnode(self.handle)
         while nh is not None:
-            yield Node(self,gv.agnameof(nh))
-            nh=gv.agnxtnode(nh)
+            yield Node(self,nh=nh)
+            nh=gv.agnxtnode(self.handle,nh)
         raise StopIteration
 
     iternodes=nodes_iter
@@ -352,19 +361,19 @@ class AGraph(object):
         """
         if v is None: (u,v)=u  # no v given, assume u is an edge tuple
         try:
-            uh=Node(self,u).get_handle()
+            uh=Node(self,u).handle
         except:
             self.add_node(u)
-            uh=Node(self,u).get_handle()
+            uh=Node(self,u).handle
         try:
-            vh=Node(self,v).get_handle()
+            vh=Node(self,v).handle
         except:
             self.add_node(v)
-            vh=Node(self,v).get_handle()
+            vh=Node(self,v).handle
         try:
-            eh=gv.agedge(uh,vh,key,_Action.create)
-            eh=Edge(self,u,v,key,eh)
-            eh.attr.update(**attr)
+            eh=gv.agedge(self.handle,uh,vh,key,_Action.create)
+            e=Edge(self,eh=eh)
+            e.attr.update(**attr)
         except KeyError:
             return None # silent failure for strict graph, already added
 
@@ -400,24 +409,28 @@ class AGraph(object):
         return Edge(self,u,v,key)
 
 
-    def delete_edge(self, u, v=None, key=None):
-        """Delete edge between nodes u and v from the graph.
+    def remove_edge(self, u, v=None, key=None):
+        """Remove edge between nodes u and v from the graph.
 
-        With optional key argument  will only delete an edge
+        With optional key argument will only remove an edge
         matching (u,v,key).
 
         """
         if v is None: (u,v)=u  # no v given, assume u is an edge tuple
         e=Edge(self,u,v,key)
         try:
-            gv.agdeledge(e.get_handle())
+            gv.agdeledge(self.handle,e.handle)
         except KeyError:
             raise KeyError("edge %s-%s not in graph"%(u,v))
 
-    def delete_edges_from(self, ebunch): 
-        """Delete edges from ebunch (a container of edges)."""
+    delete_edge=remove_edge        
+
+    def remove_edges_from(self, ebunch): 
+        """Remove edges from ebunch (a container of edges)."""
         for e in ebunch:
-            self.delete_edge(e)
+            self.remove_edge(e)
+
+    delete_edges_from=remove_edges_from
 
     def has_edge(self, u, v=None, key=None):
         """Return True an edge u-v is in the graph or False if not.
@@ -438,53 +451,7 @@ class AGraph(object):
         except KeyError:
             return False
 
-    def edges_iter(self, nbunch=None):
-        """Return iterator over edges in the graph.
-
-        If the optional nbunch (container of nodes) only edges
-        adjacent to nodes in nbunch will be returned.
-        """
-
-        if nbunch is None:   # all nodes
-            nh=gv.agfstnode(self.handle)
-            while nh is not None:
-                eh=gv.agfstout(nh)
-                while eh is not None:
-                    (s,t)=(gv.agtail(eh),gv.aghead(eh))
-                    (u,v)=(gv.agnameof(s),gv.agnameof(t))
-                    yield Edge(self,u,v,key=gv.agnameof(eh),eh=eh)
-                    eh=gv.agnxtout(eh)
-                nh=gv.agnxtnode(nh)
-        elif nbunch in self: # if nbunch is a single node 
-            n=Node(self,nbunch)
-            nh=n.get_handle()
-            eh=gv.agfstedge(nh)
-            while eh is not None:
-                (s,t)=(gv.agtail(eh),gv.aghead(eh))
-                (u,v)=(gv.agnameof(s),gv.agnameof(t))
-                yield Edge(self,u,v,key=gv.agnameof(eh),eh=eh)
-                eh=gv.agnxtedge(eh,nh)
-        else:                # if nbunch is a sequence of nodes
-            try: bunch=[n for n in nbunch if n in self]
-            except TypeError:
-                raise TypeError(
-                      "nbunch is not a node or a sequence of nodes")
-            for n in nbunch:
-                try: 
-                    nh=Node(self,n).get_handle()
-                except KeyError:
-                    continue
-                eh=gv.agfstout(nh)
-                while eh is not None:
-                    (s,t)=(gv.agtail(eh),gv.aghead(eh))
-                    (u,v)=(gv.agnameof(s),gv.agnameof(t))
-                    yield Edge(self,u,v,key=gv.agnameof(eh),eh=eh)
-                    eh=gv.agnxtout(eh)
-        raise StopIteration
- 
-    iteredges=edges_iter
-
-    def edges(self, nbunch=None):
+    def edges(self, nbunch=None, keys=False):
         """Return list of edges in the graph.
 
         If the optional nbunch (container of nodes) only edges
@@ -498,7 +465,7 @@ class AGraph(object):
         >>> print G.edges('a')
         [('a', 'b')]
         """
-        return list(self.edges_iter(nbunch))
+        return list(self.edges_iter(nbunch=nbunch,keys=keys))
 
     def has_neighbor(self, u, v, key=None):
         """Return True if u has an edge to v or False if not.
@@ -516,16 +483,15 @@ class AGraph(object):
     def neighbors_iter(self,n):
         """Return iterator over the nodes attached to n."""
         n=Node(self,n)
-        nh=n.get_handle()
-        eh=gv.agfstedge(nh)
+        nh=n.handle
+        eh=gv.agfstedge(self.handle,nh)
         while eh is not None:
-            (s,t)=(gv.agtail(eh),gv.aghead(eh))
-            (u,v)=(gv.agnameof(s),gv.agnameof(t))
-            if u==n:
-                yield Node(self,v)
+            (s,t)=Edge(self,eh=eh)
+            if s==n:
+                yield Node(self,t)
             else:
-                yield Node(self,u)
-            eh=gv.agnxtedge(eh,nh)
+                yield Node(self,s)
+            eh=gv.agnxtedge(self.handle,eh,nh)
         raise StopIteration
 
     def neighbors(self, n):
@@ -534,31 +500,36 @@ class AGraph(object):
 
     iterneighbors=neighbors_iter
 
-    def out_edges_iter(self, nbunch=None):
+    def out_edges_iter(self, nbunch=None, keys=False):
         """Return iterator over out edges in the graph.
 
         If the optional nbunch (container of nodes) only out edges
         adjacent to nodes in nbunch will be returned.
         """
+
         if nbunch is None:   # all nodes
             nh=gv.agfstnode(self.handle)
             while nh is not None:
-                eh=gv.agfstout(nh)
+                eh=gv.agfstout(self.handle,nh)
                 while eh is not None:
-                    (s,t)=(gv.agtail(eh),gv.aghead(eh))
-                    (u,v)=(gv.agnameof(s),gv.agnameof(t))
-                    yield Edge(self,u,v,key=gv.agnameof(eh),eh=eh)
-                    eh=gv.agnxtout(eh)
-                nh=gv.agnxtnode(nh)
+                    e=Edge(self,eh=eh)
+                    if keys:
+                        yield (e[0],e[1],e.name)
+                    else:
+                        yield e
+                    eh=gv.agnxtout(self.handle,eh)
+                nh=gv.agnxtnode(self.handle,nh)
         elif nbunch in self: # if nbunch is a single node 
             n=Node(self,nbunch)
-            nh=n.get_handle()
-            eh=gv.agfstout(nh)
+            nh=n.handle
+            eh=gv.agfstout(self.handle,nh)
             while eh is not None:
-                (s,t)=(gv.agtail(eh),gv.aghead(eh))
-                (u,v)=(gv.agnameof(s),gv.agnameof(t))
-                yield Edge(self,u,v,key=gv.agnameof(eh),eh=eh)
-                eh=gv.agnxtout(eh)
+                e=Edge(self,eh=eh)
+                if keys:
+                    yield (e[0],e[1],e.name)
+                else:
+                    yield e
+                eh=gv.agnxtout(self.handle,eh)
         else:                # if nbunch is a sequence of nodes
             try: bunch=[n for n in nbunch if n in self]
             except TypeError:
@@ -566,21 +537,24 @@ class AGraph(object):
                       "nbunch is not a node or a sequence of nodes")
             for n in nbunch:
                 try: 
-                    nh=Node(self,n).get_handle()
+                    nh=Node(self,n).handle
                 except KeyError:
                     continue
-                eh=gv.agfstout(nh)
+                eh=gv.agfstout(self.handle,nh)
                 while eh is not None:
-                    (s,t)=(gv.agtail(eh),gv.aghead(eh))
-                    (u,v)=(gv.agnameof(s),gv.agnameof(t))
-                    yield Edge(self,u,v,key=gv.agnameof(eh),eh=eh)
-                    eh=gv.agnxtout(eh)
+                    e=Edge(self,eh=eh)
+                    if keys:
+                        yield (e[0],e[1],e.name)
+                    else:
+                        yield e
+                    eh=gv.agnxtout(self.handle,eh)
         raise StopIteration
  
 
     iteroutedges=out_edges_iter
+    iteredges=out_edges_iter
 
-    def in_edges_iter(self, nbunch=None):
+    def in_edges_iter(self, nbunch=None, keys=False):
         """Return iterator over out edges in the graph.
 
         If the optional nbunch (container of nodes) only out edges
@@ -589,22 +563,26 @@ class AGraph(object):
         if nbunch is None:   # all nodes
             nh=gv.agfstnode(self.handle)
             while nh is not None:
-                eh=gv.agfstin(nh)
+                eh=gv.agfstin(self.handle,nh)
                 while eh is not None:
-                    (s,t)=(gv.agtail(eh),gv.aghead(eh))
-                    (u,v)=(gv.agnameof(s),gv.agnameof(t))
-                    yield Edge(self,u,v,key=gv.agnameof(eh),eh=eh)
-                    eh=gv.agnxtin(eh)
-                nh=gv.agnxtnode(nh)
+                    e=Edge(self,eh=eh)
+                    if keys:
+                        yield (e[0],e[1],e.name)
+                    else:
+                        yield e
+                    eh=gv.agnxtin(self.handle,eh)
+                nh=gv.agnxtnode(self.handle,nh)
         elif nbunch in self: # if nbunch is a single node 
             n=Node(self,nbunch)
-            nh=n.get_handle()
-            eh=gv.agfstin(nh)
+            nh=n.handle
+            eh=gv.agfstin(self.handle,nh)
             while eh is not None:
-                (s,t)=(gv.agtail(eh),gv.aghead(eh))
-                (u,v)=(gv.agnameof(s),gv.agnameof(t))
-                yield Edge(self,u,v,key=gv.agnameof(eh),eh=eh)
-                eh=gv.agnxtin(eh)
+                e=Edge(self,eh=eh)
+                if keys:
+                    yield (e[0],e[1],e.name)
+                else:
+                    yield e
+                eh=gv.agnxtin(self.handle,eh)
         else:                # if nbunch is a sequence of nodes
             try: bunch=[n for n in nbunch if n in self]
             except TypeError:
@@ -612,15 +590,17 @@ class AGraph(object):
                       "nbunch is not a node or a sequence of nodes")
             for n in nbunch:
                 try: 
-                    nh=Node(self,n).get_handle()
+                    nh=Node(self,n).handle
                 except KeyError:
                     continue
-                eh=gv.agfstin(nh)
+                eh=gv.agfstin(self.handle,nh)
                 while eh is not None:
-                    (s,t)=(gv.agtail(eh),gv.aghead(eh))
-                    (u,v)=(gv.agnameof(s),gv.agnameof(t))
-                    yield Edge(self,u,v,key=gv.agnameof(eh),eh=eh)
-                    eh=gv.agnxtin(eh)
+                    e=Edge(self,eh=eh)
+                    if keys:
+                        yield (e[0],e[1],e.name)
+                    else:
+                        yield e
+                    eh=gv.agnxtin(self.handle,eh)
         raise StopIteration
 
 
@@ -629,35 +609,34 @@ class AGraph(object):
     # define edges to be out_edges implicitly since edges uses edges_iter
     edges_iter=out_edges_iter
             
-    def out_edges(self, nbunch=None):
+    def out_edges(self, nbunch=None, keys=False):
         """Return list of out edges in the graph.
 
         If the optional nbunch (container of nodes) only out edges
         adjacent to nodes in nbunch will be returned.
         """
-        return list(self.out_edges_iter(nbunch))
+        return list(self.out_edges_iter(nbunch=nbunch,keys=keys))
 
-    def in_edges(self, nbunch=None):
+    def in_edges(self, nbunch=None, keys=False):
         """Return list of in edges in the graph.
         If the optional nbunch (container of nodes) only in edges
         adjacent to nodes in nbunch will be returned.
         """
-        return list(self.in_edges_iter(nbunch))
+        return list(self.in_edges_iter(nbunch=nbunch,keys=keys))
 
 
     def predecessors_iter(self,n):
         """Return iterator over predecessor nodes of n."""
         n=Node(self,n)
-        nh=n.get_handle()
-        eh=gv.agfstin(nh)
+        nh=n.handle
+        eh=gv.agfstin(self.handle,nh)
         while eh is not None:
-            (s,t)=(gv.agtail(eh),gv.aghead(eh))
-            (u,v)=(gv.agnameof(s),gv.agnameof(t))
-            if u==n:
-                yield Node(self,v)
+            (s,t)=Edge(self,eh=eh)
+            if s==n:
+                yield Node(self,t)
             else:
-                yield Node(self,u)
-            eh=gv.agnxtin(eh)
+                yield Node(self,s)
+            eh=gv.agnxtin(self.handle,eh)
         raise StopIteration
 
 
@@ -666,16 +645,15 @@ class AGraph(object):
     def successors_iter(self,n):
         """Return iterator over successor nodes of n."""
         n=Node(self,n)
-        nh=n.get_handle()
-        eh=gv.agfstout(nh)
+        nh=n.handle
+        eh=gv.agfstout(self.handle,nh)
         while eh is not None:
-            (s,t)=(gv.agtail(eh),gv.aghead(eh))
-            (u,v)=(gv.agnameof(s),gv.agnameof(t))
-            if u==n:
-                yield Node(self,v)
+            (s,t)=Edge(self,eh=eh)
+            if s==n:
+                yield Node(self,t)
             else:
-                yield Node(self,u)
-            eh=gv.agnxtout(eh)
+                yield Node(self,s)
+            eh=gv.agnxtout(self.handle,eh)
         raise StopIteration
 
     itersucc=successors_iter
@@ -693,12 +671,11 @@ class AGraph(object):
     out_neighbors=successors
     in_neighbors=predecessors
 
-    def degree_iter(self,nbunch=None,with_labels=False,indeg=True,outdeg=True):
+    def degree_iter(self,nbunch=None,indeg=True,outdeg=True):
         """Return an iterator over the degree of the nodes given in
-        nbunch container.
-        
-        Using optional with_labels=True returns paris of (node,degree) .
+        nbunch container.  
 
+        Returns paris of (node,degree).        
         """
         # prepare nbunch
         if nbunch is None:   # include all nodes via iterator
@@ -709,33 +686,26 @@ class AGraph(object):
             try: bunch=[Node(self,n) for n in nbunch if n in self]
             except TypeError:
                 raise TypeError("nbunch is not a node or a sequence of nodes")
-        if with_labels:
-            for n in bunch:
-                yield (Node(self,n),gv.agdegree(n.get_handle(),indeg,outdeg))
-        else:
-            for n in bunch:
-                yield gv.agdegree(n.get_handle(),indeg,outdeg)
+        for n in bunch:
+            yield (Node(self,n),gv.agdegree(self.handle,
+                                            n.handle,indeg,outdeg))
 
-
-
-    def in_degree_iter(self,nbunch=None,with_labels=False):
+    def in_degree_iter(self,nbunch=None):
         """Return an iterator over the in-degree of the nodes given in
         nbunch container.
-        
-        Using optional with_labels=True returns paris of (node,degree) .
 
+        Returns paris of (node,degree).        
         """
-        return self.degree_iter(nbunch,with_labels,indeg=True,outdeg=False)
+        return self.degree_iter(nbunch,indeg=True,outdeg=False)
 
-    def out_degree_iter(self,nbunch=None,with_labels=False):
+    def out_degree_iter(self,nbunch=None):
         """Return an iterator over the out-degree of the nodes given in
         nbunch container.
-        
-        Using optional with_labels=True returns paris of (node,degree) .
+
+        Returns paris of (node,degree).        
 
         """
-
-        return self.degree_iter(nbunch,with_labels,indeg=False,outdeg=True)
+        return self.degree_iter(nbunch,indeg=False,outdeg=True)
 
     iteroutdegree=out_degree_iter
     iterindegree=in_degree_iter
@@ -747,9 +717,9 @@ class AGraph(object):
         keyed by node with value set to the degree.
         """
         if with_labels:
-            return dict(self.out_degree_iter(nbunch,with_labels))
+            return dict(self.out_degree_iter(nbunch))
         else:
-            dlist=list(self.out_degree_iter(nbunch,with_labels))
+            dlist=list(d for n,d in self.out_degree_iter(nbunch))
             if nbunch in self:
                 return dlist[0]
             else:
@@ -766,7 +736,7 @@ class AGraph(object):
         if with_labels:
             return dict(self.in_degree_iter(nbunch,with_labels))
         else:
-            dlist=list(self.in_degree_iter(nbunch,with_labels))
+            dlist=list(d for n,d in self.in_degree_iter(nbunch))
             if nbunch in self:
                 return dlist[0]
             else:
@@ -775,9 +745,9 @@ class AGraph(object):
         
     def reverse(self):
         """Return copy of directed graph with edge directions reversed."""
-        if self.is_directed():
+        if self.directed:
             # new empty DiGraph
-            H=self.__class__(strict=self.is_strict(),directed=True) 
+            H=self.__class__(strict=self.strict,directed=True) 
             H.graph_attr.update(self.graph_attr)
             H.node_attr.update(self.node_attr)
             H.edge_attr.update(self.edge_attr)
@@ -803,9 +773,9 @@ class AGraph(object):
 
         """
         if with_labels:
-            return dict(self.degree_iter(nbunch,with_labels))
+            return dict(self.degree_iter(nbunch))
         else:
-            dlist=list(self.degree_iter(nbunch,with_labels))
+            dlist=list(d for n,d in self.degree_iter(nbunch))
             if nbunch in self:
                 return dlist[0]
             else:
@@ -829,8 +799,8 @@ class AGraph(object):
         self.graph_attr.clear()
         # now "close" existing graph and create a new graph
         name=gv.agnameof(self.handle)
-        strict=self.is_strict()
-        directed=self.is_directed()
+        strict=self.strict
+        directed=self.directed
         gv.agclose(self.handle) 
         self.handle=gv.agraphnew(name,strict,directed) 
 
@@ -886,7 +856,9 @@ class AGraph(object):
         """Return subgraph induced by nodes in nbunch.
         """
         handle=gv.agsubg(self.handle,name,_Action.create)
-        H=self.__class__(strict=self.is_strict(),directed=self.is_directed(),handle=handle,name=name,**attr)
+        H=self.__class__(strict=self.strict,
+                         directed=self.directed,
+                         handle=handle,name=name,**attr)
         if nbunch is None: return H
         # add induced subgraph on nodes in nbunch
         bunch=self.prepare_nbunch(nbunch)
@@ -898,12 +870,13 @@ class AGraph(object):
         return H
 
 
-    def delete_subgraph(self, name):
-        """Delete subgraph with given name."""  
+    def remove_subgraph(self, name):
+        """Remove subgraph with given name."""  
         handle=gv.agsubg(self.handle,name,_Action.find)
         if handle is None: raise KeyError("subgraph %s not in graph"%name)
         gv.agdelsubg(self.handle,handle)
         
+    delete_subgraph=remove_subgraph
 
     subgraph=add_subgraph
 
@@ -912,7 +885,10 @@ class AGraph(object):
         """
         handle=gv.agparent(self.handle)
         if handle is None: return None
-        H=self.__class__(strict=self.is_strict(),directed=self.is_directed(),handle=handle,name=name)
+        H=self.__class__(strict=self.strict,
+                         directed=self.directed,
+                         handle=handle,
+                         name=name)
         return H
     
     def subgraph_root(self, nbunch=None, name=None):
@@ -920,7 +896,9 @@ class AGraph(object):
         """
         handle=gv.agroot(self.handle)
         if handle is None: return None
-        H=self.__class__(strict=self.is_strict(),directed=self.is_directed(),handle=handle,name=name)
+        H=self.__class__(strict=self.strict,
+                         directed=self.directed,
+                         handle=handle,name=name)
         return H
     
     def get_subgraph(self,name):
@@ -929,14 +907,18 @@ class AGraph(object):
         """
         handle=gv.agsubg(self.handle,name,_Action.find)
         if handle is None: return None
-        H=self.__class__(strict=self.is_strict(),directed=self.is_directed(),handle=handle)
+        H=self.__class__(strict=self.strict,
+                         directed=self.directed,
+                         handle=handle)
         return H
         
     def subgraphs_iter(self):
         """Iterator over subgraphs."""
         handle=gv.agfstsubg(self.handle)
         while handle is not None:
-            yield self.__class__(strict=self.is_strict(),directed=self.is_directed(),handle=handle)
+            yield self.__class__(strict=self.strict,
+                                 directed=self.directed,
+                                 handle=handle)
             handle=gv.agnxtsubg(handle)
         raise StopIteration
 
@@ -956,7 +938,8 @@ class AGraph(object):
             return True
         else:
             return False
-        
+
+    strict=property(is_strict)        
 
     def is_directed(self):
         """Return True if graph is directed or False if not."""
@@ -965,6 +948,8 @@ class AGraph(object):
         else:
             return False
 
+    directed=property(is_directed)
+
     def is_undirected(self):
         """Return True if graph is undirected or False if not."""
         if gv.agisundirected(self.handle)==1:
@@ -974,10 +959,10 @@ class AGraph(object):
 
     def to_undirected(self):
         """Return undirected copy of graph."""
-        if self.is_undirected():
+        if not self.directed:
             return self.copy()
         else:
-            U=AGraph(strict=self.is_strict())
+            U=AGraph(strict=self.strict)
             U.graph_attr.update(self.graph_attr)
             U.node_attr.update(self.node_attr)
             U.edge_attr.update(self.edge_attr)
@@ -999,8 +984,8 @@ class AGraph(object):
         Each undirected edge u-v is represented as two directed
         edges u->v and v->u.
         """
-        if self.is_undirected():
-            D=AGraph(strict=self.is_strict(),directed=True)
+        if not self.directed:
+            D=AGraph(strict=self.strict,directed=True)
             D.graph_attr.update(self.graph_attr)
             D.node_attr.update(self.node_attr)
             D.edge_attr.update(self.edge_attr)
@@ -1081,20 +1066,17 @@ class AGraph(object):
         return self.to_string()
 
     def from_string(self,string):
-        """ 
-        Load a graph from a string in dot format. Overwrites any existing graph.
+        """Load a graph from a string in dot format.
 
+        Overwrites any existing graph.
+        
         To make a new graph from a string use
         
-        >>> s='digraph {\n1 -> 2\n;}\n'
+        >>> s='digraph {1 -> 2}'
         >>> A=AGraph()
         >>> A.from_string(s)
-
-        or
-
         >>> A=AGraph(string=s) # specify s is a string
         >>> A=AGraph(s)  # s assumed to be a string during initialization
-
         """
         from tempfile import TemporaryFile
         fh = TemporaryFile()
@@ -1135,11 +1117,18 @@ class AGraph(object):
         Use keyword args to add additional arguments to graphviz programs.
 
         """
-        import os
 
         runprog=self._get_prog(prog)
         cmd=' '.join([runprog,args])
-        child_stdin,child_stdout,child_stderr=os.popen3(cmd, 'b')
+        p = subprocess.Popen(cmd, 
+                             shell=True,
+                             stdin=subprocess.PIPE, 
+                             stdout=subprocess.PIPE, 
+                             stderr=subprocess.PIPE, 
+                             close_fds=True)
+        (child_stdin,
+         child_stdout,
+         child_stderr) = (p.stdin, p.stdout, p.stderr)
         # Use threading to avoid blocking
         data = []
         errors = []
@@ -1158,7 +1147,8 @@ class AGraph(object):
             raise IOError("".join(errors))
 
         if len(errors)>0:
-            raise IOError("".join(errors))
+            warnings.warn("".join(errors),RuntimeWarning)
+
         return "".join(data)
 
 
@@ -1234,13 +1224,13 @@ class AGraph(object):
         will be used.  
         
         Formats (not all may be available on every system depending on
-        how Graphviz was built)::
+        how Graphviz was built)
 
-            ['canon', 'cmap', 'cmapx', 'cmapx_np', 'dia', 'dot',\
-            'fig', 'gd', 'gd2', 'gif', 'hpgl', 'imap', 'imap_np',\
-            'ismap', 'jpe', 'jpeg', 'jpg', 'mif', 'mp', 'pcl', 'pdf',\
-            'pic', 'plain', 'plain-ext', 'png', 'ps', 'ps2', 'svg',\
-            'svgz', 'vml', 'vmlz', 'vrml', 'vtx', 'wbmp', 'xdot', 'xlib'])
+            'canon', 'cmap', 'cmapx', 'cmapx_np', 'dia', 'dot',
+            'fig', 'gd', 'gd2', 'gif', 'hpgl', 'imap', 'imap_np',
+            'ismap', 'jpe', 'jpeg', 'jpg', 'mif', 'mp', 'pcl', 'pdf',
+            'pic', 'plain', 'plain-ext', 'png', 'ps', 'ps2', 'svg',
+            'svgz', 'vml', 'vmlz', 'vrml', 'vtx', 'wbmp', 'xdot', 'xlib'
 
 
         If prog is not specified and the graph has positions
@@ -1274,17 +1264,6 @@ class AGraph(object):
 
         if format is None or format=='':
             format = 'dot'
-#         else:
-#             formats=dict.fromkeys(
-#                 ['canon', 'cmap', 'cmapx', 'cmapx_np', 'dia', 'dot',\
-#                      'fig', 'gd', 'gd2', 'gif', 'hpgl', 'imap', 'imap_np',\
-#                      'ismap', 'jpe', 'jpeg', 'jpg', 'mif', 'mp', 'pcl', 'pdf',\
-#                      'pic', 'plain', 'plain-ext', 'png', 'ps', 'ps2', 'svg',\
-#                      'svgz', 'vml', 'vmlz', 'vrml', 'vtx', 'wbmp', 'xdot', \
-#                      'xlib'])
-#             if not format in formats:
-#                 raise ValueError("format %s is not one of %s"%\
-#                                  (format,', '.join(formats.keys()))) 
 
         if prog is None:
             try:
@@ -1347,7 +1326,7 @@ class AGraph(object):
                  fh=os.popen("bzcat "+path) # probably not portable
             else:
                 fh = file(path,mode=mode)
-        elif hasattr(path, 'seek'):
+        elif hasattr(path, 'write'):
             fh = path
         else:
             raise TypeError('path must be a string or file handle')
@@ -1399,20 +1378,32 @@ class Node(str):
     >>> node.attr['color']='red'
 
     """ 
-    def __new__(self,graph,name):
-        n=str.__new__(self,name)
+    def __new__(self,graph,name=None,nh=None):
+        if nh is not None:
+            n=str.__new__(self,gv.agnameof(nh))
+        else:
+            n=str.__new__(self,name)
+            try:
+                nh=gv.agnode(graph.handle,n,_Action.find)
+            except KeyError:
+                raise KeyError("node %s not in graph"%n)
+            
         n.ghandle=graph.handle
-        try:
-            nh=gv.agnode(graph.handle,n,_Action.find)
-        except KeyError:
-            raise KeyError("node %s not in graph"%n)
         n.attr=ItemAttribute(nh,1)
+        n.handle=nh
         return n
 
     def get_handle(self):
         """Return pointer to graphviz node object.""" 
         return gv.agnode(self.ghandle,self,_Action.find)
         
+#    handle=property(get_handle)
+
+    def get_name(self):
+        return gv.agnameof(self.handle)
+
+    name=property(get_name)
+
 
 class Edge(tuple):
     """Edge object based on tuple.
@@ -1448,44 +1439,36 @@ class Edge(tuple):
     >>> edge.attr['color']='red'
 
     """ 
-    def __new__(self,graph,source,target,key=None,eh=None):
-        s=Node(graph,source)
-        t=Node(graph,target)
-
-        try:
-            if eh is None:
-                eh=gv.agedge(s.get_handle(),t.get_handle(),key,_Action.find)
-        except KeyError:
-            raise KeyError("edge %s-%s not in graph"%(source,target))
-
-        if key is None:
-            tp=tuple.__new__(self,(s,t))
+    def __new__(self,graph,source=None,target=None,key=None,eh=None):
+        # edge handle given, reconstruct node object 
+        if eh is not None:
+            (source,target)=(gv.agtail(eh),gv.aghead(eh))
+            s=Node(graph,nh=source)
+            t=Node(graph,nh=target)
+        # no edge handle, search for edge and construct object             
         else:
-            tp=tuple.__new__(self,(s,t,key))
+            s=Node(graph,source)
+            t=Node(graph,target)
+            try:
+                eh=gv.agedge(graph.handle,
+                             s.handle,
+                             t.handle,
+                             key,
+                             _Action.find)
+            except KeyError:
+                raise KeyError("edge %s-%s not in graph"%(source,target))
 
+        tp=tuple.__new__(self,(s,t))
         tp.ghandle=graph.handle
+        tp.handle=eh
         tp.attr=ItemAttribute(eh,3)
+        tp.key=key
         return tp
 
-    def get_handle(self):
-        """Return pointer to graphviz edge object.""" 
-        try:
-            sh=gv.agnode(self.ghandle,self[0],_Action.find)
-        except KeyError:
-            raise KeyError("node %s not in graph"%source)
-        try:
-            th=gv.agnode(self.ghandle,self[1],_Action.find)
-        except KeyError:
-            raise KeyError("node %s not in graph"%target)
-        try:
-            if len(self)==3:
-                key=self[2]
-            else:
-                key=None
-            return gv.agedge(sh,th,key,_Action.find)
-        except KeyError:
-            raise KeyError("edge %s-%s not in graph"%(source,target))
+    def get_name(self):
+        return gv.agnameof(self.handle)
 
+    name=property(get_name)
 
 
 class Attribute(UserDict.DictMixin):
@@ -1531,11 +1514,11 @@ class Attribute(UserDict.DictMixin):
 
 
     def __getitem__(self, name):
-        try:
-            ah=gv.agattr(self.handle,self.type,name,None)
-            return gv.agattrdefval(ah)
-        except KeyError:
-            raise KeyError("no attribute found")
+        item=gv.agget(self.handle,name)
+        if item is None:
+            ah=gv.agattr(self.handle,self.type,name,None)            
+            item=gv.agattrdefval(ah)
+        return item
 
     def __delitem__(self, name):
         gv.agattr(self.handle,self.type,name,'')
@@ -1630,6 +1613,7 @@ def _test_suite():
     suite = doctest.DocFileSuite('tests/graph.txt',
                                  'tests/attributes.txt',
                                  'tests/layout_draw.txt',
+#                                 'tests/subgraph.txt',
                                  package='pygraphviz')
     doctest.testmod() # test docstrings in module
     return suite
